import typing

import pydantic
import pydantic.json_schema
import pydantic_core.core_schema
import returns.maybe
import typing_extensions


T = typing.TypeVar("T")


class _MaybeAnnotation(typing.Generic[T]):
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: typing.Any,
        handler: pydantic.GetCoreSchemaHandler,
    ) -> pydantic_core.core_schema.CoreSchema:
        origin = typing_extensions.get_origin(source_type)
        if origin is None:
            origin = source_type
            inner_type = typing.Any
        else:
            inner_type = typing_extensions.get_args(source_type)[0]
        inner_schema = handler.generate_schema(inner_type)
        from_optional_schema = pydantic_core.core_schema.chain_schema(
            [
                pydantic_core.core_schema.union_schema(
                    [
                        pydantic_core.core_schema.none_schema(),
                        inner_schema,
                    ]
                ),
                pydantic_core.core_schema.no_info_plain_validator_function(
                    returns.maybe.Maybe.from_optional
                ),
            ]
        )
        return pydantic_core.core_schema.json_or_python_schema(
            json_schema=from_optional_schema,
            python_schema=pydantic_core.core_schema.union_schema(
                [
                    pydantic_core.core_schema.is_instance_schema(returns.maybe.Maybe),
                    from_optional_schema,
                ],
            ),
            serialization=pydantic_core.core_schema.plain_serializer_function_ser_schema(
                lambda instance: instance.value_or(None)
            ),
        )


Maybe = typing.Annotated[returns.maybe.Maybe[T], _MaybeAnnotation]


class Model(pydantic.BaseModel):
    maybeInt: Maybe[int]
    maybeStr: Maybe[str]


instance = Model(maybeInt=3, maybeStr="abc")
